IRDS_TO_TIREX_DATASET = {
    'clueweb09/en/trec-web-2009': 'clueweb09-en-trec-web-2009-20230107-training',
    'clueweb09/en/trec-web-2010': 'clueweb09-en-trec-web-2010-20230107-training',
    'clueweb09/en/trec-web-2011': 'clueweb09-en-trec-web-2011-20230107-training',
    'clueweb09/en/trec-web-2012': 'clueweb09-en-trec-web-2012-20230107-training',
    'clueweb12/trec-web-2013': 'clueweb12-trec-web-2013-20230107-training',
    'clueweb12/trec-web-2014': 'clueweb12-trec-web-2014-20230107-training',
    'disks45/nocr/trec-robust-2004': 'disks45-nocr-trec-robust-2004-20230209-training',
    'antique/test': 'antique-test-20230107-training',
    'argsme/2020-04-01/touche-2020-task-1': 'argsme-touche-2020-task-1-20230209-training',
    'argsme/2020-04-01/touche-2021-task-1': 'argsme-touche-2021-task-1-20230209-training',
    'clueweb12/touche-2020-task-2': 'clueweb12-touche-2020-task-2-20230209-training',
    'clueweb12/touche-2021-task-2': 'clueweb12-touche-2021-task-2-20230209-training',
    'clueweb12/b13/trec-misinfo-2019': 'clueweb12-trec-misinfo-2019-20240214-training',
    'cord19/fulltext/trec-covid': 'cord19-fulltext-trec-covid-20230107-training',
    'cranfield': 'cranfield-20230107-training',
    'disks45/nocr/trec7': 'disks45-nocr-trec7-20230209-training',
    'disks45/nocr/trec8': 'disks45-nocr-trec8-20230209-training',
    'gov/trec-web-2002': 'gov-trec-web-2002-20230209-training',
    'gov/trec-web-2003': 'gov-trec-web-2003-20230209-training',
    'gov/trec-web-2004': 'gov-trec-web-2004-20230209-training',
    'gov2/trec-tb-2004': 'gov2-trec-tb-2004-20230209-training',
    'gov2/trec-tb-2005': 'gov2-trec-tb-2005-20230209-training',
    'gov2/trec-tb-2006': 'gov2-trec-tb-2006-20230209-training',
    'medline/2004/trec-genomics-2004': 'medline-2004-trec-genomics-2004-20230107-training',
    'medline/2004/trec-genomics-2005': 'medline-2004-trec-genomics-2005-20230107-training',
    'medline/2017/trec-pm-2017': 'medline-2017-trec-pm-2017-20230211-training',
    'medline/2017/trec-pm-2018': 'medline-2017-trec-pm-2018-20230211-training',
    'msmarco-passage/trec-dl-2019/judged': 'msmarco-passage-trec-dl-2019-judged-20230107-training',
    'msmarco-passage/trec-dl-2020/judged': 'msmarco-passage-trec-dl-2020-judged-20230107-training',
    'nfcorpus/test': 'nfcorpus-test-20230107-training',
    'vaswani': 'vaswani-20230107-training',
    'wapo/v2/trec-core-2018': 'wapo-v2-trec-core-2018-20230107-training',
    'longeval/heldout': 'longeval-heldout-20230513-training',
    'longeval/long-september': 'longeval-long-september-20230513-training',
    'longeval/short-july': 'longeval-short-july-20230513-training',
    'longeval/train': 'longeval-train-20230513-training',
    'trec-tip-of-the-tongue/dev': 'trec-tip-of-the-tongue-dev-20230607-training',
}

TIREX_DATASETS = IRDS_TO_TIREX_DATASET.values()
